/**
 * Assignment #9: POSIX select() and Socket Connection (Client)
 * Sample Solution
 *
 * Instructor: Ron Mak
 * CMPE 142 Operating Systems
 * Computer Engineering Department
 * San Jose State University
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include "globals.h"

int connect_to_socket(char *ip_address);
void print_buffer(char *buffer, int size);

/**
 * The parent process.
 * @param fd_1 the file descriptors for pipe #1
 * @param fd_1 the file descriptors for pipe #2
 * @param ip_address the IP address for the socket connection.
 */
int parent(int fd_1[], int fd_2[], char *ip_address)
{
    // Close the unused WRITE ends of the pipes.
    close(fd_1[WRITE_FD]);
    close(fd_2[WRITE_FD]);

    // 10 second timeout for select().
    struct timeval timeout;
    timeout.tv_sec  = 10;
    timeout.tv_usec = 0;

    int socket_fd = -1;

    // Infinite loop to read from all the input sources.
    while (1)
    {
        // Create a new socket connection if we
        // had read from the previous connection.
        if (socket_fd < 0) socket_fd = connect_to_socket(ip_address);

        // Determine the maximum file descriptor.
        int max_fd = socket_fd;
        if (fd_1[READ_FD] > max_fd) max_fd = fd_1[READ_FD];
        if (fd_2[READ_FD] > max_fd) max_fd = fd_2[READ_FD];

        char buffer[BUFFER_SIZE];
        memset(buffer, '0', sizeof(buffer));

        // Initialize the read_set to monitor the input sources.
        // The select() call will modify the set,
        // so we must re-initialize it every time.
        fd_set read_set;
        FD_ZERO(&read_set);                // initialize to the empty set
        FD_SET(socket_fd,     &read_set);  // set socket descriptor
        FD_SET(fd_1[READ_FD], &read_set);  // set pipe #1 READ end descriptor
        FD_SET(fd_2[READ_FD], &read_set);  // set pipe #2 READ end descriptor

        int count = select(max_fd + 1, &read_set,
                           (fd_set *) 0, (fd_set *) 0, &timeout);

        switch (count)
        {
            // Timed out with nothing ready.
            case 0:
            {
                printf("select() timed out\n");
                break;
            }

            // Error.
            case -1:
            {
                perror("*** Error: select()");
                return -1;
            }

            // Some input is ready.
            default:
            {
                // Check the socket connection.
                if (FD_ISSET(socket_fd, &read_set))
                {
                    // Read from the socket connection.
                    int size = read(socket_fd, buffer, sizeof(buffer)-1);
                    if (size > 0)
                    {
                        printf("Received from time server:");
                        print_buffer(buffer, size);
                    }
                    else perror("*** Error: read() socket");

                    close(socket_fd);
                    socket_fd = -1;
                }

                // Check pipe #1.
                if (FD_ISSET(fd_1[READ_FD], &read_set))
                {
                    int size = read(fd_1[READ_FD], buffer, sizeof(buffer)-1);
                    if (size > 0)
                    {
                        printf("Received from child 1:");
                        print_buffer(buffer, size);
                    }
                    else perror("*** Error: read() child #1");
                }

                // Check pipe #2.
                if (FD_ISSET(fd_2[READ_FD], &read_set))
                {
                    int size = read(fd_2[READ_FD], buffer, sizeof(buffer)-1);
                    if (size > 0)
                    {
                        printf("Received from child 2:");
                        print_buffer(buffer, size);
                    }
                    else perror("*** Error: read() child #2");
                }

                break;
            }
        }
    }

    return 0;
}

/**
 * Print the contents of the input buffer.
 * @param buffer the input buffer.
 * @param size the buffer size.
 */
void print_buffer(char *buffer, int size)
{
    int index = 0;

    do
    {
        printf(" %s", buffer + index);
        fflush(stdout);

        index += strlen(buffer + index) + 1;
    } while (index < size);

    printf("\n");
    fflush(stdout);
}
